/* 
 * PROGRAM:	gamon 
 * AUTHOR:	Gary Artim ( gartim@gmail.com )
 * VERSION:	0.01
 * Copyright:	"Copyright 2008 Gary Artim"
 * License:	GNU GENERAL PUBLIC LICENSE Version 2
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * DESCRIPTION:	monitor program to test active ports and connectivity.
 * 		using http/smtp/icmp protocals.
 * REQUIRED:
 * libxml2 libraries
 * emerge dev-libs/libxml2 (to install on a gentoo linux system)
 *
 * COMPILE:
 * gcc `xml2-config --cflags --libs` -o gamon gamon.c && chmod 4711 gamon
 * chown root.root gamon && chmod 4711 gamon
 *
 * FORMAT:	
 * indent -kr gamon.c
 *
 * RUNNING:
 * ./gamon --config gamon.conf --loop 8
 * ./gamon --help ( help options )
 * ./gamon --xml --config gamon.conf --loop 8 
 */
#include <stdio.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/param.h>
#include <sys/file.h>
#include <arpa/inet.h>
#include <errno.h>
#define MESSAGE_SIZE 	40
#define	MAXWAIT 	3	/* max time to wait for response, sec. */
#define	MAXPACKET	4096	/* max packet size */
#define VERBOSE		1	/* verbose flag */
#define QUIET		2	/* quiet flag */
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	64
#endif
u_char packet[MAXPACKET];
int ident, datalen, s, i, pingflags, options;
extern int errno;
struct hostent *hp;		/* Pointer to host info */
struct timezone tz;		/* leftover */
struct sockaddr whereto;	/* Who to ping */
char usage2[] =
    /* "Usage:  ping [-dfqrv] host [packetsize [count [preload]]]\n"; */
    "Usage:  ping [-dfqrv] host [packetsize [count [preload]]]\n";
char *hostname;
char hnamebuf[MAXHOSTNAMELEN];
int npackets = 1;
int raw = 0;
int xml = 0;
int loop = 0;
int nodenum = 1;
unsigned looptime = 0;
int preload = 0;		/* number of packets to "preload" */
int ntransmitted = 0;		/* sequence # for outbound packets = #sent */
int nreceived = 0;		/* # of packets we got back */
int timing = 1;
int tmin = 999999999;
int tmax = 0;
int tsum = 0;			/* sum of all times, for doing average */
int cnt = 0, gaveup = 0;
void catcher();
char *inet_ntoa();
xmlDocPtr doc = NULL;		/* document pointer */
xmlNodePtr root_node = NULL, node = NULL, node1 = NULL;	// node pointers 
xmlDtdPtr dtd = NULL;		/* DTD pointer */
char buff[256];
struct timezone tz;
struct host {
    char name[256];
    struct host *next;
};
struct netprobe {
    struct host *icmphead, *icmptail, *httphead, *httptail, *smtphead,
	*smtptail;
};
struct netprobe np;
static int debug_flag;

void timesub(endtime, starttime)
register struct timeval *endtime, *starttime;
{
    if ((endtime->tv_usec -= starttime->tv_usec) < 0) {
	endtime->tv_sec--;
	endtime->tv_usec += 1000000;
    }
    endtime->tv_sec -= starttime->tv_sec;
}
int hostlist(char *type, struct host *p)
{
    while (p->next != NULL) {
	printf("type %s host [%s]\n", type, p->next->name);
	p = p->next;
    }
}
int ctlparse(char *file)
{
    char ch, *lbuff, buff[256];
    int I, len, line = 0, cnt = 0, EOL = 10;
    lbuff = buff;
    I = open(file, O_RDONLY);
    /* remove EOL and "=" with NULLs to create 2 strings */
    while (read(I, &ch, 1) == 1) {
	if (ch == '#' && cnt == 0) {
	    while (read(I, &ch, 1) == 1) {
		if (ch == EOL || ch == EOF)
		    break;
	    }
	} else if (ch == ' ') {
	    /* skip spaces */
	} else {
	    *lbuff = ch;
	    if (*lbuff == '=')
		*lbuff = (char) NULL;
	    if (*lbuff == EOL && cnt > 0) {
		*lbuff = (char) NULL;
		cnt = 0;
		lbuff = buff;
		++line;
		if (debug_flag) {
		    printf("s1: (%s)\n", buff);
		    printf("l1: (%d)\n", strlen(buff));
		    printf("s2: (%s)\n", buff + (strlen(buff)) + 1);
		    printf("s2: (%d)\n",
			   strlen(buff + (strlen(buff)) + 1));
		    printf("line %d\n", line);
		}
		proctoken(buff, buff + (strlen(buff)) + 1);
	    } else {
		++cnt;
		++lbuff;
		if (cnt > 255) {
		    printf("gamon: config opt error at line %d\n", ++line);
		    exit(4);
		}
	    }
	}
    }
}
int proctoken(char *l, char *r)
{
    struct host *p = (struct host *) malloc(sizeof(struct host));
    strcpy(p->name, r);
    p->next = NULL;
    if (strcmp(l, "icmp") == 0) {
	if (np.icmphead == np.icmptail)
	    /* first entry */
	    np.icmphead->next = p;
	np.icmptail->next = p;
	np.icmptail = p;
    } else if (strcmp(l, "http") == 0) {
	if (np.httphead == np.httptail)
	    /* first entry */
	    np.httphead->next = p;
	np.httptail->next = p;
	np.httptail = p;
    } else if (strcmp(l, "smtp") == 0) {
	if (np.smtphead == np.smtptail)
	    /* first entry */
	    np.smtphead->next = p;
	np.smtptail->next = p;
	np.smtptail = p;
    } else {
	printf("gamon: error, invalid .conf file, exiting\n");
	exit(4);
    }
}
int usage()
{
    printf("usage: gamon [options]\n");
    printf("              --config  s config file\n");
    printf("              --debug     produce debug output\n");
    printf("              --help      this help output\n");
    printf("              --loop    n loop every n seconds\n");
    printf("              --raw       output raw results\n");
    printf("              --xml       XML output\n");
    exit;
};
int gopt(int argc, char **argv)
{
    int c;
    while (1) {
	static struct option long_options[] = {
	    {"debug", no_argument, &debug_flag, 1},
	    {"config", required_argument, 0, 'c'},
	    {"help", no_argument, 0, 'h'},
	    {"loop", required_argument, 0, 'l'},
	    {"raw", no_argument, 0, 'r'},
	    {"xml", no_argument, 0, 'x'},
	    {0, 0, 0, 0}
	};
	int option_index = 0;
	c = getopt_long(argc, argv, "hlp:c:", long_options, &option_index);
	if (c == -1)
	    break;
	switch (c) {
	case 0:
	    if (long_options[option_index].flag != 0)
		break;
	    break;
	case 'c':
	    ctlparse(optarg);
	    break;
	case 'h':
	    printf("option --help\n");
	    usage();
	    break;
	case 'l':
	    /* looptime = atoi(optarg); */
	    if (!sscanf(optarg, "%2d", &looptime)) {
		printf("gamon: exit, invalid --loop seconds\n");
		exit(1);
	    }
	    loop = 1;
	    break;
	case 'r':
	    raw = 1;
	    break;
	case 'x':
	    xml = 1;
	    break;
	case '?':
	    break;
	default:
	    abort();
	}
    }
    return (0);
}
void giveup(int sig)
{
    gaveup = 1;
}
int tcp_connect(char *host, char *service)
{
    struct hostent *host_info;
    struct servent *serv_info;
    struct sigaction act;
    struct sockaddr_in server = { 0 };
    int sock, rc;
    act.sa_handler = giveup;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGALRM, &act, 0);
    sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    host_info = gethostbyname(host);
    if (!host_info) {
	printf("hostname problem for [%s] protocal [%s], exiting\n", host,
	       service);
	exit(1);
    }
    serv_info = getservbyname(service, "tcp");
    server.sin_family = AF_INET;
    memcpy((char *) &server.sin_addr, host_info->h_addr,
	   host_info->h_length);
    server.sin_port = serv_info->s_port;
    alarm(4);
    rc = connect(sock, (struct sockaddr *) &server, sizeof(server));
    alarm(0);			/* need to clear alarm is successful */
    if (gaveup) {
	rc = -1;
	gaveup = 0;
    }
    if (rc != -1) {
	return sock;
    } else {
	return rc;
    }
}
int checkhosts(char *type, struct host *p, char *string)
{
    int fdnet, status, ok = 0, rc;
    time_t starttime, endtime;
    time_t *stime = &starttime;
    time_t *etime = &endtime;
    double timeittook = 0;
    char result[2048];
    char outstring[256];
    struct timeval tv_start, tv_end;
    int triptime;
    while (p->next != NULL) {
	if (raw)
	    printf("[\n[ host (%30s) service (%4s)\n[ rawmode\n[\n[\n",
		   p->next->name, type);
	if (strcmp(type, "icmp") == 0) {
	    icmptest(p->next->name);
	} else {
	    ok = 0;
	    fdnet = tcp_connect(p->next->name, type);
	    if (fdnet == -1) {
		if (!raw) {
		    if (!xml) {
			printf
			    ("host [%50s] time [  -1ms] service [%4s] sockdown\n",
			     p->next->name, type);
		    } else {
			sprintf(outstring, "%50s", p->next->name);
			sprintf(buff, "host");
			node =
			    xmlNewChild(root_node, NULL, BAD_CAST buff,
					BAD_CAST outstring);
			sprintf(buff, "%4s", type);
			xmlNewProp(node, BAD_CAST "type", BAD_CAST buff);
			xmlNewProp(node, BAD_CAST "timing",
				   BAD_CAST "  -1");
			xmlNewProp(node, BAD_CAST "status", BAD_CAST "sd");
		    }
		} else {
		    printf
			("gamon: host %30s service %4s, cannot creat socket\n",
			 p->next->name, type);
		}
	    } else {
		gettimeofday(&tv_start, &tz);
		if (strcmp(type, "http") == 0 || strcmp(type, "smtp") == 0) {
		    rc = write(fdnet, string, strlen(string));
		}
		bzero(result, sizeof(result));
		while ((status =
			read(fdnet, result, sizeof(result) - 1)) > 0) {
		    if (raw) {
			printf("%d: %s", status, result);
		    } else {
			if (strcmp(type, "smtp") == 0) {
			    if (strstr(result, "220"))
				ok = 1;
			}
			if (strcmp(type, "http") == 0) {
			    if (strstr(result, "HTTP")
				&& (strstr(result, "200")
				    || strstr(result, "301")))
				ok = 1;
			}
		    }
		}
		gettimeofday(&tv_end, &tz);
		timesub(&tv_end, &tv_start);
		triptime = tv_end.tv_sec * 1000 + (tv_end.tv_usec / 1000);

		if ((!raw) && ok) {
		    if (!xml) {
			printf
			    ("host [%50s] time [%4dms] service [%4s] up\n",
			     p->next->name, triptime, type);
		    } else {
			sprintf(outstring, "%50s", p->next->name);
			sprintf(buff, "host");
			node =
			    xmlNewChild(root_node, NULL, BAD_CAST buff,
					BAD_CAST outstring);
			sprintf(buff, "%4s", type);
			xmlNewProp(node, BAD_CAST "type", BAD_CAST buff);
			sprintf(buff, "%4d", triptime);
			xmlNewProp(node, BAD_CAST "timing", BAD_CAST buff);
			xmlNewProp(node, BAD_CAST "status", BAD_CAST "up");
		    }
		} else if (!raw) {
		    if (!xml) {
			printf
			    ("host [%50s] time [%4dms] service [%4s] down\n",
			     p->next->name, triptime, type);
		    } else {
			sprintf(outstring, "%50s", p->next->name);
			sprintf(buff, "host");
			node =
			    xmlNewChild(root_node, NULL, BAD_CAST buff,
					BAD_CAST outstring);
			sprintf(buff, "%4s", type);
			xmlNewProp(node, BAD_CAST "type", BAD_CAST buff);
			xmlNewProp(node, BAD_CAST "timing", BAD_CAST "-1");
			xmlNewProp(node, BAD_CAST "status",
				   BAD_CAST "down");
		    }
		}
		close(fdnet);
	    }
	}
	p = p->next;
	if (raw)
	    printf("]\n]\n]\n]\n]\n\n\n");
    }
}
icmptest(char *hostname)
{
    struct sigaction act;
    struct sockaddr_in from;
    struct sockaddr_in *to = (struct sockaddr_in *) &whereto;
    int on = 1;
    struct protoent *proto;
    bzero((char *) &whereto, sizeof(struct sockaddr));
    to->sin_family = AF_INET;
    hp = gethostbyname(hostname);
    if (hp) {
	to->sin_family = hp->h_addrtype;
	bcopy(hp->h_addr, (caddr_t) & to->sin_addr, hp->h_length);
	hostname = hp->h_name;
    } else {
	printf("gamon:: icmp:: unknown host %s\n", hostname);
	exit(1);
    }
    datalen = 64 - 8;
    ident = getpid() & 0xFFFF;
    if ((proto = getprotobyname("icmp")) == NULL) {
	fprintf(stderr, "gamon:: icmp:: unknown protocol\n");
	exit(10);
    }
    if ((s = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
	perror("gamon:: icmp:: socket");
	exit(5);
    }
    if (raw) {
	if (to->sin_family == AF_INET) {
	    printf("PING %s (%s): %d data bytes\n", hostname,
		   inet_ntoa(to->sin_addr), datalen);
	} else {
	    printf("PING %s: %d data bytes\n", hostname, datalen);
	}
    }
    setlinebuf(stdout);
    act.sa_handler = catcher;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGALRM, &act, 0);
    /* fire off them quickies */
    for (i = 0; i < preload; i++)
	pinger();
    catcher();			/* start things going */
    for (;;) {
	int len = sizeof(packet);
	int fromlen = sizeof(from);
	int cc;
	struct timeval timeout;
	int fdmask = 1 << s;
	timeout.tv_sec = 0;
	timeout.tv_usec = 10000;
	if ((cc =
	     recvfrom(s, packet, len, 0, (struct sockaddr *) &from,
		      &fromlen)) < 0) {
	    if (errno == EINTR)
		continue;
	    perror("gamon:: icmp:: recvfrom");
	    continue;
	}
	pr_pack(packet, cc, &from, hostname);
	if (npackets && nreceived >= npackets) {
	    return (0);
	}
    }
}
void catcher()
{
    int waittime;
    pinger();
    if (npackets == 0 || ntransmitted < npackets)
	alarm(1);
    else {
	if (nreceived) {
	    waittime = 2 * tmax / 1000;
	    if (waittime == 0)
		waittime = 1;
	} else
	    waittime = MAXWAIT;
	alarm(waittime);
    }
}

pinger()
{
    static u_char outpack[MAXPACKET];
    register struct icmp *icp = (struct icmp *) outpack;
    int i, cc;
    register struct timeval *tp = (struct timeval *) &outpack[8];
    register u_char *datap = &outpack[8 + sizeof(struct timeval)];
    icp->icmp_type = ICMP_ECHO;
    icp->icmp_code = 0;
    icp->icmp_cksum = 0;
    icp->icmp_seq = ntransmitted++;
    icp->icmp_id = ident;	/* ID */
    cc = datalen + 8;		/* skips ICMP portion */
    if (timing)
	gettimeofday(tp, &tz);
    for (i = 8; i < datalen; i++)	/* skip 8 for time */
	*datap++ = i;
    icp->icmp_cksum = in_cksum(icp, cc);
    i = sendto(s, outpack, cc, 0, &whereto, sizeof(struct sockaddr));
    if (i < 0 || i != cc) {
	if (i < 0)
	    perror("sendto");
	if (raw) {
	    printf("ping: wrote %s %d chars, ret=%d\n", hostname, cc, i);
	} else {
	    printf("host [%50s] service [icmp] up\n", hostname);
	}
	fflush(stdout);
    }
}
char *pr_type(t)
register int t;
{
    static char *ttab[] = {
	"Echo Reply",
	"ICMP 1",
	"ICMP 2",
	"Dest Unreachable",
	"Source Quench",
	"Redirect",
	"ICMP 6",
	"ICMP 7",
	"Echo",
	"ICMP 9",
	"ICMP 10",
	"Time Exceeded",
	"Parameter Problem",
	"Timestamp",
	"Timestamp Reply",
	"Info Request",
	"Info Reply"
    };
    if (t < 0 || t > 16)
	return ("OUT-OF-RANGE");
    return (ttab[t]);
}

pr_pack(buf, cc, from, hostname)
char *buf, *hostname;
int cc;
struct sockaddr_in *from;
{
    struct ip *ip;
    struct sockaddr_in cvt;
    register struct icmp *icp;
    register long *lp = (long *) packet;
    register int i;
    struct timeval tv;
    struct timeval *tp;
    int hlen, triptime, testit;
    char outstring[256];
    from->sin_addr.s_addr = ntohl(from->sin_addr.s_addr);
    gettimeofday(&tv, &tz);
    ip = (struct ip *) buf;
    hlen = ip->ip_hl << 2;
    if (cc < hlen + ICMP_MINLEN) {
	if (pingflags & VERBOSE)
	    printf("packet too short (%d bytes) from %s\n", cc,
		   inet_ntoa(from->sin_addr));
	return;
    }
    cc -= hlen;
    icp = (struct icmp *) (buf + hlen);
    if ((!(pingflags & QUIET)) && icp->icmp_type != ICMP_ECHOREPLY) {
	cvt.sin_addr.s_addr = ntohl(from->sin_addr.s_addr);
	printf("%d bytes from %s: icmp_type=%d (%s) icmp_code=%d\n",
	       cc, inet_ntoa(cvt.sin_addr),
	       icp->icmp_type, pr_type(icp->icmp_type), icp->icmp_code);
	if (pingflags & VERBOSE) {
	    for (i = 0; i < 12; i++)
		printf("x%2.2x: x%8.8x\n", i * sizeof(long), *lp++);
	}
	return;
    }
    if (icp->icmp_id != ident)
	return;			/* 'Twas not our ECHO */
    if (timing) {
	tp = (struct timeval *) &icp->icmp_data[0];
	/* tvsub( &tv, tp ); */
	timesub(&tv, tp);
	triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);
	tsum += triptime;
	if (triptime < tmin)
	    tmin = triptime;
	if (triptime > tmax)
	    tmax = triptime;
    }
    if (!(pingflags & QUIET)) {
	cvt.sin_addr.s_addr = ntohl(from->sin_addr.s_addr);
	if (raw) {
	    printf("%d bytes from %s: icmp_seq=%d", cc, inet_ntoa(cvt.sin_addr), icp->icmp_seq);	/* DFM */
	    if (timing) {
		printf(" time=%d ms\n", triptime);
	    } else {
		putchar('\n');
	    }
	} else {
	    if (!xml) {
		printf("host [%50s] time [%4dms] service [%4s] up\n",
		       hostname, triptime, "icmp");
	    } else {
		sprintf(outstring, "%50s", hostname);
		sprintf(buff, "host");
		node = xmlNewChild(root_node, NULL, BAD_CAST buff,
				   BAD_CAST outstring);
		sprintf(buff, "%4d", triptime);
		xmlNewProp(node, BAD_CAST "type", BAD_CAST "icmp");
		xmlNewProp(node, BAD_CAST "timing", BAD_CAST buff);
		xmlNewProp(node, BAD_CAST "status", BAD_CAST "up");
	    }
	}
    }
    nreceived++;
}

in_cksum(addr, len)
u_short *addr;
int len;
{
    register int nleft = len;
    register u_short *w = addr;
    register u_short answer;
    register int sum = 0;
    /*
     *  Our algorithm is simple, using a 32 bit accumulator (sum),
     *  we add sequential 16 bit words to it, and at the end, fold
     *  back all the carry bits from the top 16 bits into the lower
     *  16 bits.
     */
    while (nleft > 1) {
	sum += *w++;
	nleft -= 2;
    }
    if (nleft == 1) {
	u_short u = 0;

	*(u_char *) (&u) = *(u_char *) w;
	sum += u;
    }
    /*
     * add back carry outs from top 16 bits to low 16 bits
     */
    sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
    sum += (sum >> 16);		/* add carry */
    answer = ~sum;		/* truncate to 16 bits */
    return (answer);
}


int main(int argc, char **argv)
{
    int fdnet, ret, newuid;
    struct host *picmp = (struct host *) malloc(sizeof(struct host));
    struct host *phttp = (struct host *) malloc(sizeof(struct host));
    struct host *psmtp = (struct host *) malloc(sizeof(struct host));
    np.icmphead = np.icmptail = picmp;
    np.httphead = np.httptail = phttp;
    np.smtphead = np.smtptail = psmtp;
    np.icmphead->next = np.httphead->next = np.smtphead->next = NULL;
    gopt(argc, argv);
    if (debug_flag) {
	hostlist("icmp", np.icmphead);
	hostlist("http", np.httphead);
	hostlist("smtp", np.smtphead);
    }
    do {
	if ((!raw) && xml) {
	    LIBXML_TEST_VERSION;
	    doc = xmlNewDoc(BAD_CAST "1.0");
	    root_node = xmlNewNode(NULL, BAD_CAST "root");
	    xmlDocSetRootElement(doc, root_node);
	    dtd =
		xmlCreateIntSubset(doc, BAD_CAST "root", NULL,
				   BAD_CAST "tree2.dtd");
	}
	checkhosts("http", np.httphead, "HEAD / HTTP/1.0\n\n");
	checkhosts("smtp", np.smtphead, "RSET\nquit\n");
	checkhosts("icmp", np.icmphead, "");
	alarm(0);		/* clear any outstanding alarms */
	if ((!raw) && xml) {
	    xmlSaveFormatFileEnc("-", doc, "UTF-8", 1);
	    xmlFreeDoc(doc);
	    /* Free the global variables that may */
	    /* have been allocated by the parser. */
	    xmlCleanupParser();
	    /* this is to debug memory for regression tests */
	    xmlMemoryDump();
	}
	sleep(looptime);
    } while (loop);
    return (0);
}
